import subprocess
import os, time, traceback
import threading
import logging

from collections import namedtuple
from queue import Queue

try:
    import prettytable as pt
except ModuleNotFoundError:
    from utils import prettytable as pt

                
GPUInfo = namedtuple('GPUInfo', ('id','memory_total', 'memory_free', 'memory_used'))



def _get_gpu_info():
    gpu_info = os.popen('nvidia-smi --query-gpu=index,memory.total,memory.free,memory.used --format=csv')
    gpu_info.readline()

    gpu_data = []
    for line in gpu_info:
        index, memory_total, memory_free, memory_used = line.strip().split(', ')
        index = int(index)
        memory_total = float(memory_total.split()[0])
        memory_free = float(memory_free.split()[0])
        memory_used = float(memory_used.split()[0])
        gpu_data.append(GPUInfo(index, memory_total, memory_free, memory_used))
    return gpu_data


def _get_use_gpu(allow_free=None):
    gpu_data = get_gpu_info()
    if allow_free is None:
        return max(gpu_data, key=lambda x: x.memory_free)
    else:
        return list(filter(lambda x: x.memory_free > allow_free, gpu_data))


def get_gpu_info_test():
    gpu_data = []
    with open('./gpulimit/test/Resources', 'r') as f:
        for line in f:
            id, free = line.strip().split(':')
            id = int(id)
            free = float(free)
            gpu_data.append(GPUInfo(id, 0, free, 0))
    return gpu_data


def get_use_gpu_test(allow_free=None):
    gpu_data = get_gpu_info()
    if allow_free is None:
        return max(gpu_data, key=lambda x: x.memory_free)
    else:
        return list(filter(lambda x: x.memory_free > allow_free, gpu_data))


if not os.environ.get('GPULIMIT_DEBUG'):
    try:
        result = _get_gpu_info()
    except Exception:
        result = None
    if not result:
        print('[Warning]: can not use `nvidia-smi`, please check cuda environment.')
        print('[info]: set GPULIMIT_DEBUG=1.')
        os.environ['GPULIMIT_DEBUG'] = '1'

if os.environ.get('GPULIMIT_DEBUG'):
    logging.warning('here use debug environment. NO GPU USE!!!')
    get_gpu_info = get_gpu_info_test
    get_use_gpu = get_use_gpu_test
else:
    get_gpu_info = _get_gpu_info
    get_use_gpu = _get_use_gpu



        

class TaskStatus(object):
    status2id = dict(zip(['CMD_ERROR', 'complete', 'waiting', 'running', 'runtime_error', 'killed'], range(-1, 5)))
    id2status = dict(zip(status2id.values(), status2id.keys()))
    
    can_start_list = ['waiting', 'runtime_error', 'killed']
    auto_start_list = ['waiting', 'runtime_error']
    
    start_sort_type = {
        'waiting': 0,
        'runtime_error': 1,
        'killed': 2,
        'running': 3,
        'complete': 3,
        'CMD_ERROR': 3,
    }
    show_sort_type = {
        'running': 1,
        'waiting': 2,
        'killed': 3,
        'runtime_error': 3,
        'complete': 4,
        'CMD_ERROR': 5,
    }
    
    def __init__(self, status='waiting', run_times=0, err_code=None):
        self._status = None
        self.run_times = run_times
        self.err_code = err_code
        
        if isinstance(status, int):
            self._set_status_id(status)
            
        elif isinstance(status, str):
            self._set_status_name(status)       
            
        else:
            raise TypeError(f'status must be int or str, not {type(status)}')
    
    @property
    def id(self):
        return self.status2id[self.status]
    
    @id.setter
    def id(self, nid):
        self._set_status_id(nid)
        
    @property
    def status(self):
        return self._status
    
    @status.setter
    def status(self, new_status):
        self._set_status_name(new_status)
        
    def _set_status_id(self, id):
            if id in self.id2status:
                self._status = self.id2status[id]
            else:
                raise ValueError(f'status id must be: {self.id2status}')
    
    def _set_status_name(self, name):
            if name in self.status2id:
                self._status = name
            else:
                raise ValueError(f'status name must be: {self.status2id}')
            
    def __str__(self):
        return self.status
    
    def __repr__(self):
        return f'TaskStatus({self.status})'


class Task(object):
    def __init__(self, task_queue, id, pwd, cmds, out_path=None):
        self.task_queue = task_queue
        self.id = id
        self.pwd = pwd
        self.cmds = cmds
        self.out_path = out_path
        
        self.gpu = None
        self.is_in_queue = False
        
        self.run_times = 0
        self.out_file = None
        self.pkg_process = None
        self.process = None
        self.available = True

        self.killed = False
        self.debug_msg = None
        
    def _run_task(self, GPU_id):
        if self.out_path is not None:
            self.out_file = open(self.out_path, 'w')
        self.gpu = GPU_id
        env = os.environ.copy()
        env['CUDA_VISIBLE_DEVICES'] = str(GPU_id)
        try:
            self.process = subprocess.Popen(self.cmds, stdout=self.out_file, shell=True, stderr=subprocess.STDOUT, cwd=self.pwd, env=env)
            logging.info(f'[starting({self.id}: GPU:{GPU_id})]: {self.pwd}$ {self.cmds}')
        except Exception as e:
            self.available = False
            logging.info(f'[CMD_ERROR({self.id}: GPU:{GPU_id})]: {self.pwd}$ {self.cmds} \nerror:\n{e}')
            msg = traceback.format_exc()
            logging.info(msg)
            self.debug_msg = 'cmds: ' + str(self.cmds) + "\n" + msg
            self.task_queue.check_and_start()
            return
        self.process.wait()
        self.gpu = None
        self.is_in_queue = False
        logging.info(f'[finish({self.id}: GPU:{GPU_id})]: {self.pwd}$ {self.cmds}')
        
        self.task_queue.check_and_start()

    def start(self, GPU_id):
        if self.killed:
            self.killed = False
        if self.status.status in ['waiting', 'runtime_error', 'killed']:
            self.run_times += 1
            self.pkg_process = threading.Thread(target=self._run_task, args=(GPU_id,))
            self.pkg_process.start()
            
            return 0, f'[info]: start task {self.id} succeed.'
        else:
            return 1, f'[info]: can not start task {self.id} which have status `{self.status.status}`'
        
    def kill(self):
        if self.process is not None and self.status.status == 'running':
            self.process.terminate()
            time.sleep(1)
            self.process.kill()
            self.killed = True
            
            self.gpu = None
            self.is_in_queue = False
            return 0, f'[info]: kill task {self.id} succeed.'
        else:
            return 1, f'[warning]: can not kill task {self.id} which have status `{self.status.status}`'

    @property
    def status(self):
        if not self.available:
            return TaskStatus('CMD_ERROR', self.run_times)
        
        if self.process is None:
            return TaskStatus('waiting', self.run_times)
        
        if self.killed:
            return TaskStatus('killed', self.run_times)
        
        status = self.process.poll()
        if status is None:
            return TaskStatus('running', self.run_times)
        
        if status == 0:
            return TaskStatus('complete', self.run_times)
        else:
            return TaskStatus('runtime_error', self.run_times, err_code=status)


#def client(client_cmd):
#    def decorator(func):
#        def wrapper(self, *args, **kwargs):
#            self.func_map[client_cmd] = func
#            return func(*args, **kwargs)
#        return wrapper
#    return decorator


class TaskQueue(object):
    def __init__(self, logdir='./tmp', MINI_MEM_REMAIN=1024, MAX_ERR_TIMES=5, WAIT_TIME=10):
        self.queue = []
        self.id_give = 0
        self.logdir = logdir
        self.log_file = os.path.join(self.logdir, 'main.log')
        
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        else:
            # del logfiles in log dir
            for logname in os.listdir(self.logdir):
                path = os.path.join(logdir, logname) 
                if os.path.isfile(path):
                    os.remove(path)

        if os.environ.get('GPULIMIT_DEBUG'):
            logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(message)s')
            logging.basicConfig(level=logging.WARNING, format='%(asctime)s - %(message)s')
        else:
            logging.basicConfig(filename=self.log_file, level=logging.INFO, format='%(asctime)s - %(message)s')
            logging.basicConfig(filename=self.log_file, level=logging.WARNING, format='%(asctime)s - %(message)s')

#        self.func_map = {}
        
        self.MINI_MEM_REMAIN = MINI_MEM_REMAIN
        self.MAX_ERR_TIMES = MAX_ERR_TIMES
        self.WAIT_TIME = WAIT_TIME
        
        self.start_thread = threading.Thread(target=self._thread_start_task)
        self.need_start_tasks = Queue()
        self.lock = threading.RLock()
        self.start_thread.start()
        
    
    def _thread_start_task(self):
        while True:
            if not self.need_start_tasks.empty():
                task = self.need_start_tasks.get()
                if task.status.status in TaskStatus.can_start_list:
                    self.run_task(task)
                    time.sleep(self.WAIT_TIME)
            else:
                time.sleep(1)
    
    @staticmethod
    def _check_input(input_types, extra_args=(), extra_kwargs={}):
        '''
        Example:
            
            input_types = (('1', int), ('start', str))
        '''
        err_msg = ''
        result_input = []
        for value, dtype in input_types:
            try:
                result_input.append(dtype(value))
            except Exception as e:
                result_input.append(None)
                err_msg += f'[error]: input {value} is not type `{dtype.__name__}`.\n'
        for v in extra_args:
            err_msg += f'[error]: can not identify param `{v}`.\n'
        for k, v in extra_kwargs.items():
            err_msg += f'[error]: can not identify param `{k}={v}`.\n'
            
        return result_input, err_msg
    
    def add(self, pwd, cmds, logpath=None):
        if logpath is None:
            logpath = os.path.join(self.logdir, f'{self.id_give}.log')
        task = Task(self, self.id_give, pwd, cmds, logpath)
        self.id_give += 1
        self.queue.append(task)
        result = f'add task(id:{task.id}) to queue(len: {len(self.queue)})'
        logging.info(f'add task(id:{task.id}): {pwd}$ {cmds})')
        err, result_ = self.check_and_start()
        return err, '\n'.join([result, result_])

    def run_task(self, task):
        use_gpu = get_use_gpu()
        errcode, result = task.start(use_gpu.id)   
        return errcode, result

#    @client('ls')
    def ls(self, *, all=False, sort='show'):
        '''
        ls                            ls GPU task queue status
        '''
        (all, sort), err_msg = self._check_input(((all, bool), (sort,str)))
        if err_msg: return 1, err_msg
        
        err_code, err_msg = self._sort_priority(sort)
        if err_msg: return 1, err_msg
        
        result = f'TaskQueue MINI_MEM_REMAIN={self.MINI_MEM_REMAIN}, MAX_ERR_TIMES={self.MAX_ERR_TIMES}\n'
        table = pt.PrettyTable(['[ID]', 'num', 'status', 'run_times', 'pwd', 'cmds'])
        table.border = False
#        result += '[ID]\tnum\t|\tstatus\trun_times\tcmds\n'
        for i, task in enumerate(self.queue):
            status = str(task.status) + f'(GPU:{task.gpu})' if task.gpu is not None else str(task.status)
            if not all:
                table.add_row([task.id, i, status, task.run_times, task.pwd+'#', " ".join(task.cmds)[:80]])
            else:
                table.add_row([task.id, i, status, task.run_times, task.pwd+'#', " ".join(task.cmds)])
#            result += f'{task.id}\t{i}\t|\t{str(task.status)}\t{task.run_times}\t{task.pwd}# {"".join(task.cmds)}\n'
        return 0, result + str(table)

    def rm(self, id):
        '''
        rm [id]                       remove task [id] from manage, if task is running, kill it.
        '''
        (id,), err_msg = self._check_input(((id, int),))
        if err_msg:
            return 1, err_msg
        
        result = ''
        self.lock.acquire()
        for i, task in enumerate(self.queue):
            if task.id == id:
                if str(self.queue[i].status) == 'running':
                    self.queue[i].kill()
                del self.queue[i]
                result += f'[info]: del task {id}'
                break
        self.lock.release()
        if not result:
            return 1, f'[error]: can not found {id} in task queue.'
        else:
            return 0, result

    def kill(self, id):
        '''
        kill [id]                     kill task [id]
        '''
        
        (id,), err_msg = self._check_input(((id, int),))
        if err_msg:
            return 1, err_msg
        
        for i, task in enumerate(self.queue):
            if task.id == id:
                return task.kill()
        return 1, f'[error]: can not found id {id} in task queue.'
        
    def move_to_top(self, id, index=0, *args, **kwargs):
        '''
        move [id] [index(default=0)]  move [id] to [index]
        '''
        (id, index), err_msg = self._check_input(((id, int),(index, int)), args, kwargs)
        if err_msg:
            return 1, err_msg
        
        if index > len(self.queue):
            return 2, f'[error]: index {index} is bigger than task queue length({len(self.queue)})'
        
        pos = None
        for i, task in enumerate(self.queue):
            if task.id == id:
                pos = i
                break
        if pos is not None:
            self.lock.acquire()
            t = self.queue[pos]
            del self.queue[pos]
            self.queue.insert(index, t)
            self.lock.release()
            return 0, f'[info]: move {id} to the first'
        else:
            return 1, f'[error]: can not found task {id}'


    def _sort_priority(self, type='start'):
        if type == 'start':
            sort_type = TaskStatus.start_sort_type
        elif type == 'show':
            sort_type = TaskStatus.show_sort_type
        elif type == 'id':
            sort_type = dict(zip(TaskStatus.status2id.keys(), [1] * len(TaskStatus.status2id)))
        else:
            return 1, f'[Error]: can not found sort type `{type}`, which can use `show` `id` `start`'
            
        self.lock.acquire()
        self.queue = sorted(self.queue, key=lambda x: x.id)
        self.queue = sorted(self.queue, key=lambda x: sort_type[x.status.status])
        self.lock.release()
        return 0, ''

    def check_and_start(self): 
        result = ''
        can_use_gpu = get_use_gpu(self.MINI_MEM_REMAIN)
        if can_use_gpu:
            self._sort_priority('start')
            remain_tasks = list(filter(lambda task: task.status.status in TaskStatus.auto_start_list and not task.is_in_queue, self.queue))
            if remain_tasks:
                self.need_start_tasks.put(remain_tasks[0])
                remain_tasks[0].is_in_queue = True
            else:
                result = f'all task is to be run or completed.'
        else:
            result = 'GPU memory is full. task is waitting for others.'
            
        logging.info(result)
        return 0, result

    def clean(self, *args):
        '''
        clean [type(default=None)]    remove complete task and CMD_ERROR task.
        
        Example:
            
            gpulimit clean            clean all `CMD_ERROR` `complete` status task
            gpulimit clean kill       clean all `kill` status task
        '''
        if not args:
            rm_types = ['CMD_ERROR', 'complete']
        else:
            rm_types = list(args)
            
        rm_ids = []
        rm_cmds = []
        rm_pwds = []
        for task in filter(lambda x:x.status.status in rm_types, self.queue):
            rm_ids.append(task.id)
            rm_cmds.append(task.cmds)
            rm_pwds.append(task.pwd)
        self.lock.acquire()
        self.queue = list(filter(lambda x:x, [None if task.status.status in rm_types else task for task in self.queue]))
        self.lock.release()
        return 0, '\n'.join([f'[info]: rm ID({id_})  {pwd}# {cmds}' for id_, pwd, cmds in zip(rm_ids, rm_pwds, rm_cmds)])

    def set_property(self, name, value):
        '''
        set [name] [value]            set some property.
        
        Example:
            
            gpulimit set WAIT_TIME 1  set `WAIT_TIME=1`
        '''
        
        name2type = {
            'MINI_MEM_REMAIN': int, 
            'MAX_ERR_TIMES': int,
            'WAIT_TIME': int,
        }
        if name in name2type:
            value = name2type[name](value)
            self.__dict__[name] = value
            result = 0, f'[info]: set {name}={value}'
        else:
            result = 1, f'[error]: name `{name}` can not set.'
        logging.info(result[1])
        return result
    
    def start(self, id=None):
        '''
        start [iddefalut=None]        Force start task(s).
        
        Information:
            
            gpulimit start            running `check_and_start`, and auto start new task.
            gpulimit start 1          Force start task [id].
        '''
        
        if id is None:
            return self.check_and_start()
        (id, ), err_msg = self._check_input(((id, int),), )
        if err_msg:
            return 1, err_msg
        
        for task in self.queue:
            if task.id == id:
                return self.run_task(task)

        return 1, f'[error]: can not found task[{id}]'
                
    def get_output_filename(self, id):
        '''
        log [id]                      show [id] output.
        
        Example:
            
            gpulimit log 1            show task(id=1) output.
            gpulimit log main         show manage background log info.
        '''
        
        if id=='main':
            return 0, self.log_file
        
        (id, ), err_msg = self._check_input(((id, int),), )
        if err_msg:
            return 1, err_msg
        
        path = ''
        for i, task in enumerate(self.queue):
            if id == task.id:
                path = os.path.abspath(task.out_path)
        if path:
            return 0, path
        else:
            return 1, 'Error'
        
    def status(self):
        '''
        status                        show GPU status.
        
        Example:
            
            Nothing
        '''
        gpu_data = get_gpu_info()
        task_nums = [0] * len(gpu_data)
        for task in self.queue:
            if task.gpu:
                task_nums[task.gpu] += 1
                
        table = pt.PrettyTable(['GPU[ID]', 'memory total', 'memory free', 'memory used', 'running tasks num'])
        table.border = False
        for info, use_num in zip(gpu_data, task_nums):
            table.add_row([info.id, info.memory_total, info.memory_free, info.memory_used, use_num])
            
        return 0, str(table)
    
    def debug(self, id):
        '''
        debug [id]                    if task [id] is `CMD_ERROR`, use this show error traceback.
        
        Example:
            
            debug 1                   show task 1 error traceback.                
        '''
        (id, ), err_msg = self._check_input(((id, int),), )
        if err_msg:
            return 1, err_msg
        
        for task in self.queue:
            if task.id == id:
                return 0, str(task.debug_msg)
        return 1, f'[error]: can not found task[{id}]'


#! /usr/bin/python3
import os, sys, inspect
import time
import socket
import traceback

import pickle as pk

parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__))) 
sys.path.insert(0, parentdir) 

from gpulimit.gpulimit_core import recv_all, send_all_str
from gpulimit.gpulimit_core import task_manage


"""
GPU Task Manage
"""


class Server(object):
    def __init__(self):
        if sys.platform == 'linux':
            server_address = '/tmp/gpulimit_uds_socket'
            try:
                os.unlink(server_address)
            except OSError:
                if os.path.exists(server_address):
                    raise RuntimeError('server is running.')
        else:
            server_address = ('0.0.0.0', 5123)
        self.server_address = server_address
        
        self.task_manage = task_manage
        if sys.platform == 'linux':
            self.task_manage.start(logdir='/tmp/')
        else:
            self.task_manage.start(logdir='./tmp/')
        
        self.func_map = {
                
            '-h': self._help,
            '--help': self._help,
            'help': self._help, 
            
            'add': self.task_manage.add,
            
        }
#        print(self.task_manage.func_map)
        self.func_map.update(self.task_manage.func_map)
        
        
    def start(self):
        if isinstance(self.server_address, str):
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(self.server_address)
        sock.listen(5)
        print('start gpulimit server.')
        print(f'listening at {self.server_address}')
        while True:
            connection, client_address = sock.accept()
            self._process(connection)
            connection.close()
            
    def _process(self, sock):
        msgs = recv_all(sock)
        pwd, cmds = pk.loads(msgs)
        
        try:
            if cmds[0] == 'add':
                
                msg = self._create_task(pwd, cmds)
            else:
                msg = self._process_commands(pwd, cmds)
        except Exception:
            msg = traceback.format_exc()
            
        send_all_str(sock, msg)  
        
    def _get_args(self, cmds, get_all=True):
        add_arg = True
        
        args = []
        kwargs = {}
        for cmd in cmds:
            if cmd.startswith('--'):
                if not get_all:
                    add_arg = False
                    
                cmd = cmd.lstrip('--')
                splits = cmd.split('=')
                if len(splits) == 1:
                    kwargs[cmd] = True
                    continue
                key = splits[0]
                value = '='.join(splits[1:])
                kwargs[key] = value
            else:
                if add_arg:
                    args.append(cmd)
                    
        return args, kwargs
        
    def _check_input(self, func, args, kwargs):
        fullargspec = inspect.getfullargspec(func)
        return_msg = ''
        if fullargspec.varkw is None:
            for key in kwargs:
                if not key in fullargspec.kwonlyargs:
                    return_msg += f'[Error]: not support param `{key}`. \n'
        if fullargspec.varargs is None:
            if len(fullargspec.args) == 0:
                max_args_len = 0
            else:
                max_args_len = len(fullargspec.args)-1 if fullargspec.args[0] == 'self' else len(fullargspec.args)
            if max_args_len < len(args):
                return_msg += f'[Error]: have max {max_args_len} input, but you input {len(args)} args. \n'
        return return_msg
    
        
    def _create_task(self, pwd, cmds):
        _, kwargs = self._get_args(cmds[1:], False)
        i = -1
        for i, cmd in enumerate(cmds[1:]):
            if not cmd.startswith('--'):
                break
        cmds = cmds[i + 1:]
        if len(cmds) == 0:
            return f'[Error]: you input args {kwargs}, but no cmd input.'
        err_msg = self._check_input(self.task_manage.add, (), kwargs)
        if err_msg: return err_msg
        _, result = self.task_manage.add(pwd, cmds, **kwargs)
        return result
        
    def _process_commands(self, pwd, cmds):
        if self.func_map.get(cmds[0]):
            func = self.func_map[cmds[0]]
            
            args, kwargs = self._get_args(cmds[1:])         
            err_msg = self._check_input(func, args, kwargs)
            if err_msg: return err_msg
            
            _, return_msg = func(*args, **kwargs)

        else:
            return_msg = '[error]: no cmd found.'
        return return_msg
        
    def _help(self, cmd=None):
        '''
        help [cmd]                    show help
        '''
        
        header = '''
GPU Task Manage:
    usage:
        
        client.py -h                  show help
        gpulimit add [cmds]           add task [cmds] to gpulimit queue.
        

    other commands:\n'''
    
        if cmd is not None:
            if self.func_map.get(cmd):
                func = self.func_map.get(cmd)
                if func.__doc__ is not None:
                    return 0, func.__doc__
                else:
                    return 0, f'\n        {cmd}\t\tno info\n'
            else:
                return 1, f'[Error]: Can not found command: `{cmd}`.\n Please use help to find more infomation.'
        
        func_info = ''
        for cmd, func in self.func_map.items():
            if not cmd.startswith('-'):
                if func.__doc__ is not None:
                    func_info += list(filter(lambda x: x, func.__doc__.split('\n')))[0] + '\n'
                else:
                    func_info += f'        {cmd}\t\tno info\n'

        return 0, header + func_info
    

def main():
    server = Server()
    server.start()
    

if __name__=='__main__':
    main()



# -*- coding: utf-8 -*-


def check_input(input_types, extra_args=(), extra_kwargs={}):
    '''
    Example:
        
        input_types = (('1', int), ('start', str))
    '''
    err_msg = ''
    result_input = []
    for value, dtype in input_types:
        try:
            result_input.append(dtype(value))
        except Exception as e:
            result_input.append(None)
            err_msg += f'[error]: input {value} is not type `{dtype.__name__}`.\n'
    for v in extra_args:
        err_msg += f'[error]: can not identify param `{v}`.\n'
    for k, v in extra_kwargs.items():
        err_msg += f'[error]: can not identify param `{k}={v}`.\n'
            
    return result_input, err_msg




import abc
import logging

from .system_info import system_info
from .tasks import TaskStatus

class Scheduling(abc.ABC):
    
    @abc.abstractmethod
    def callback_process_end(self, task_manage, *args, **kwargs):
        pass
    
    @abc.abstractmethod
    def callback_add_process(self, task_manage, *args, **kwargs):
        pass
    
    @abc.abstractmethod
    def timer_call(self, task_manage, *args, **kwargs):
        pass
    
    @abc.abstractmethod
    def user_start_scheduling(self, task_manage, task_id=None):
        pass
    
    
class BaseScheduling(Scheduling):
    def __init__(self):
        pass
    
    @staticmethod
    def _use_gpu_id(info):
        return sorted(info.gpu, key = lambda x: x.memory_free, reverse=True)[0].id
        
    def callback_process_end(self, task_manage, *args, **kwargs):
        pass
    
    def callback_add_process(self, task_manage, *args, **kwargs):
        return 0, ''
    
    def timer_call(self, task_manage, *args, **kwargs):
        info = system_info.refresh()
        if len(info.gpu) == 0:
            for task in task_manage.tasks:
                if task.run_times > task_manage.setter_param['MAX_ERR_TIMES']:
                    continue
                if task.status.status in TaskStatus.auto_start_list:
                    task.start(self._use_gpu_id(info))
                    logging.info(f'start task {task.id}.')
                    return True
            return False
        
        if info.gpu[self._use_gpu_id(info)].memory_free > task_manage.setter_param['MINI_MEM_REMAIN']:
            for task in task_manage.tasks:
                if task.run_times > task_manage.setter_param['MAX_ERR_TIMES']:
                    continue
                if task.status.status in TaskStatus.auto_start_list:
                    task.start(self._use_gpu_id(info))
                    logging.info(f'start task {task.id}.')
                    return True
        return False
    
    def user_start_scheduling(self, task_manage, task_id=None):
        if task_id is None:
            return self.callback_add_process(task_manage)
        
        task = task_manage.get_task(task_id)
        info = system_info.refresh()
        
        logging.info(f'start task {task.id}.')
        return task.start(self._use_gpu_id(info))


#! /usr/bin/python3
import os, sys, inspect
import time
import socket
import traceback

import pickle as pk

parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__))) 
sys.path.insert(0, parentdir) 

from gpulimit.gpulimit_core.socket_utils import recv_all, send_all_str
from gpulimit.gpulimit_core.task_mutiprocess import TaskQueue


"""
GPU Task Manage
"""


class Server(object):
    def __init__(self):
        if sys.platform == 'linux':
            server_address = '/tmp/gpulimit_uds_socket'
            try:
                os.unlink(server_address)
            except OSError:
                if os.path.exists(server_address):
                    raise RuntimeError('server is running.')
        else:
            server_address = ('0.0.0.0', 5123)
        self.server_address = server_address
        
        if sys.platform == 'linux':
            self.task_queue = TaskQueue(logdir='/tmp/', MINI_MEM_REMAIN=1024)
        else:
            self.task_queue = TaskQueue(logdir='./tmp/', MINI_MEM_REMAIN=1024)
        
#        print('task_queue.func_map:', self.task_queue.func_map)
        self.func_map = {
                
            '-h': self._help,
            '--help': self._help,
            'help': self._help, 
            
            'ls': self.task_queue.ls,
            
            'kill': self.task_queue.kill, 
            
            'start': self.task_queue.start,
            
            'rm': self.task_queue.rm, 
    
            'move': self.task_queue.move_to_top, 
    
            'log': self.task_queue.get_output_filename, 
    
            'clean': self.task_queue.clean, 
            
            'set': self.task_queue.set_property, 
            
            'status': self.task_queue.status,
            
            'debug': self.task_queue.debug,
            
        }
    def start(self):
        if isinstance(self.server_address, str):
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(self.server_address)
        sock.listen(5)
        print('start gpulimit server.')
        print(f'listening at {self.server_address}')
        while True:
            connection, client_address = sock.accept()
            self._process(connection)
            connection.close()
            
    def _process(self, sock):
        msgs = recv_all(sock)
        pwd, cmds = pk.loads(msgs)
        
        try:
            if cmds[0] == 'add':
                msg = self._create_task(pwd, cmds[1:])
            else:
                msg = self._process_commands(pwd, cmds)
        except Exception:
            msg = traceback.format_exc()
            
        send_all_str(sock, msg)
            
    def _create_task(self, pwd, cmds):
        _, result = self.task_queue.add(pwd, cmds)
        return result
        
    def _process_commands(self, pwd, cmds):
        if self.func_map.get(cmds[0]):
            func = self.func_map[cmds[0]]
            
            fullargspec = inspect.getfullargspec(func)
                
            args = []
            kwargs = {}
            for cmd in cmds[1:]:
                if cmd.startswith('--'):
                    cmd = cmd.lstrip('--')
                    splits = cmd.split('=')
                    if len(splits) == 1:
                        kwargs[cmd] = True
                        continue
                    key = splits[0]
                    value = '='.join(splits[1:])
                    kwargs[key] = value
                else:
                    args.append(cmd)
            
            return_msg = ''
            if fullargspec.varkw is None:
                for key in kwargs:
                    if not key in fullargspec.kwonlyargs:
                        return_msg += f'[Error]: {cmds[0]} not support param `{key}`. \n'
            if fullargspec.varargs is None:
                if len(fullargspec.args) - 1 < len(args):
                    return_msg += f'[Error]: {cmds[0]} have max {len(fullargspec.args) - 1} input, but you input {len(args)} args. \n'
            if return_msg: return return_msg
            
            _, return_msg = func(*args, **kwargs)

        else:
            return_msg = '[error]: no cmd found.'
        return return_msg
        
    def _help(self, cmd=None):
        '''
        help [cmd]                    show help
        '''
        
        header = '''
GPU Task Manage:
    usage:
        
        client.py -h                  show help
        gpulimit add [cmds]           add task [cmds] to gpulimit queue.
        

    other commands:\n'''
    
        if cmd is not None:
            if self.func_map.get(cmd):
                func = self.func_map.get(cmd)
                if func.__doc__ is not None:
                    return 0, func.__doc__
                else:
                    return 0, f'\n        {cmd}\t\tno info\n'
            else:
                return 1, f'[Error]: Can not found command: `{cmd}`.\n Please use help to find more infomation.'
        
        func_info = ''
        for cmd, func in self.func_map.items():
            if not cmd.startswith('-'):
                if func.__doc__ is not None:
                    func_info += list(filter(lambda x: x, func.__doc__.split('\n')))[0] + '\n'
                else:
                    func_info += f'        {cmd}\t\tno info\n'

        return 0, header + func_info
    

def main():
    server = Server()
    server.start()
    

if __name__=='__main__':
    main()



import socket
import struct


def send_all(sock, msg):
    sock.send(struct.pack('>Q', len(msg)))
    sock.sendall(msg)
    
def recv_all(sock, buffer_size=1024):
    msg_len = sock.recv(8)
    msg_len = struct.unpack('>Q', msg_len)[0]
    msg = b''
    while len(msg) < msg_len:
        msg += sock.recv(buffer_size)
    return msg
    

def send_all_str(sock, msg):
    msg = bytes(msg, encoding='utf8')
    send_all(sock, msg)
    
def recv_all_str(sock, buffer_size=1024):
    msg = recv_all(sock, buffer_size)
    msg = msg.decode('utf8')
    return msg
#%%
    
#sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#sock.bind(('0.0.0.0', 19999))
#sock.listen(1)
#
#conn,addr = sock.accept() 
#print(recv_all(conn))
#send_all_str(conn, 'dqwwwdwd')
#sock.close()

#%%

#sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#sock.connect(('127.0.0.1', 19999))
#send_all_str(sock, 'awsefwseefeewfdsifdslxc')
#print(recv_all_str(sock))



import os, sys, time, logging
from collections import deque, namedtuple

SystemInfo = namedtuple('SystemInfo', ('time', 'CPU_utilization','memory', 'gpu', 'task'))
MemoryInfo = namedtuple('MemoryInfo', ('total','free', 'used'))
GPUInfo = namedtuple('GPUInfo', ('id','memory_total', 'memory_free', 'memory_used', 'utilization'))
ProcessInfo = namedtuple('ProcessInfo', ('pid','memory_used', 'gpu_memory'))

class System(object):
    def __init__(self, debug=False):
        self.data = deque(maxlen=1)
        self.debug = debug
    
    @property
    def debug(self):
        return self._debug
    @debug.setter
    def debug(self, value):
        self._debug = bool(value)
        if self._debug:
            self.get_gpu_info = self._get_gpu_info_test
        else:
            self.get_gpu_info = self._get_gpu_info
        try:
            self.get_gpu_info()
        except Exception:
            self.get_gpu_info = lambda: None
            logging.warn('Can not use `nvidia-smi`, please check you `PATH` environment.')
            
    def refresh(self):
        self.data.append(SystemInfo(time.time(), self.get_cpu_utilization(), 
                                    self.get_memory_info(), self.get_gpu_info(), 
                                    self.get_pid_info()))
        return self
    
    def __call__(self):
        self.refresh()
        return self.data[-1]
    
    @property
    def gpu(self):
        if len(self.data) == 0:
            raise RuntimeError('call `refresh` first.')
        return self.data[-1].gpu
    
    @property
    def pid(self):
        if len(self.data) == 0:
            raise RuntimeError('call `refresh` first.')
        return self.data[-1].task
    
    @staticmethod
    def _get_gpu_info():
        gpu_info = os.popen('nvidia-smi --query-gpu=index,memory.total,memory.free,memory.used,utilization.gpu --format=csv')
        gpu_info.readline()
    
        gpu_data = []
        for line in gpu_info:
            index, memory_total, memory_free, memory_used, utilization = line.strip().split(', ')
            index = int(index)
            memory_total = float(memory_total.split()[0])
            memory_free = float(memory_free.split()[0])
            memory_used = float(memory_used.split()[0])
            utilization = float(utilization.split()[0])
            gpu_data.append(GPUInfo(index, memory_total, memory_free, memory_used, utilization))
        return gpu_data
    
    @staticmethod
    def _get_gpu_info_test():
        gpu_data = []
        with open('./gpulimit/test/Resources', 'r') as f:
            for line in f:
                id, free = line.strip().split(':')
                id = int(id)
                free = float(free)
                gpu_data.append(GPUInfo(id, 0, free, 0, 0))
        return gpu_data
    
    @staticmethod
    def get_memory_info():
        if sys.platform == 'linux': 
            total, free = os.popen('cat /proc/meminfo').readlines()[:2]
            total = float(total.strip().split(':')[1].split()[0])
            free = float(free.strip().split(':')[1].split()[0])
        else:
            free = float(os.popen('wmic OS get FreePhysicalMemory').readlines()[2].strip())
            total = float(os.popen('wmic ComputerSystem get TotalPhysicalMemory').readlines()[2].strip())
        used = total - free
        return MemoryInfo(total, free, used)
    
    @staticmethod
    def get_pid_info():
#        data = {}
#        if sys.platform == 'linux': 
#            pass
#        else:
#            for line in os.popen('TASKLIST /FO CSV').readlines()[1:]:
#                print(line)
#                name, pid, _, _, memory_used = line.strip().split(',')
#                pid = int(pid.strip('"'))
#                memory_used = memory_used.strip('"').split()[0]
#                memory_used = ''.join(memory_used.split(','))
#                try:
#                    memory_used = float(memory_used)
#                except Exception:
#                    memory_used = 0
#                data[pid] = ProcessInfo(pid, memory_used, None)
        return {}
            
    @staticmethod
    def get_cpu_utilization():
        return None
    

system_info = System()
if system_info.get_gpu_info() is None:
    system_info.debug = True



# -*- coding: utf-8 -*-
import os, time
import threading
import logging
import heapq # TODO

from gpulimit.utils import prettytable as pt
from gpulimit.utils import check_input

from .system_info import system_info
from .tasks import Sort, Task
from .scheduling import BaseScheduling


class TaskManage(object):
    def __init__(self, scheduling):
        self.queue = []
        self._id_give = 0
        
        self.logdir = None
        self.log_file = None
        
        if os.environ.get('GPULIMIT_DEBUG'):
            logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(message)s')
            logging.basicConfig(level=logging.WARNING, format='%(asctime)s - %(message)s')
        else:
            logging.basicConfig(filename=self.log_file, level=logging.INFO, format='%(asctime)s - %(message)s')
            logging.basicConfig(filename=self.log_file, level=logging.WARNING, format='%(asctime)s - %(message)s')

        self.func_map = {}
        self.scheduling = scheduling
        
        self.setter_param = {
            'MINI_MEM_REMAIN': 1024,
            'MAX_ERR_TIMES': 5,
            'WAIT_TIME': 10,
        }
        
        self.start_thread = threading.Thread(target=self._thread_start_task)
        self.lock = threading.RLock()
        
        
    def start(self, logdir='./tmp', MINI_MEM_REMAIN=1024, MAX_ERR_TIMES=5, WAIT_TIME=10):
        self.logdir = logdir
        self.log_file = os.path.join(self.logdir, 'main.log')
        
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        else:
            # del logfiles in log dir
            for logname in os.listdir(self.logdir):
                path = os.path.join(logdir, logname) 
                if os.path.isfile(path):
                    os.remove(path)
                    
        self.setter_param = {
            'MINI_MEM_REMAIN': MINI_MEM_REMAIN,
            'MAX_ERR_TIMES': MAX_ERR_TIMES,
            'WAIT_TIME': WAIT_TIME,
        }
        
        self.start_thread.start()
        
        
    def _thread_start_task(self):
        while True:
            result = self.scheduling.timer_call(self)
            if result:
                time.sleep(self.setter_param['WAIT_TIME'])
            else:
                time.sleep(1)
    
    @property
    def tasks(self):
        return self.queue.copy()
    
    def __len__(self):
        return len(self.queue)
    
    def add_task(self, new_task, priority=5):
        new_task.priority = priority
        i = 0
        for i, task in enumerate(self.queue):
            if priority >= task.priority:
                i = i + 1
            if priority < task.priority:
                break
            
        self.lock.acquire()
        self.queue.insert(i, new_task)
        self.lock.release()
    
    def get_task(self, id):
        for task in self.queue:
            if task.id == id:
                return task
        return None
    
    def rm_task(self, id):
        task = self.get_task(id)
        if task is None:
            return False
        task.kill()
        self.lock.acquire()
        self.queue.remove(task)
        self.lock.release()
        return True
           
    def mv_task(self, id, index):
        task = self.get_task(id)
        if task is None:
            return False
        self.lock.acquire()
        self.queue.remove(task)
        self.queue.insert(index, task)
        self.lock.release()
        return True
    
    def change_priority(self, id, priority):
        task = self.get_task(id)
        if task is None:
            return False
        self.lock.acquire()
        self.queue.remove(task)
        self.lock.release()
        self.add_task(task, priority)
        return True
    
    def client(self, client_cmd):
        def decorator(func):
            self.func_map[client_cmd] = func
            return func
        return decorator
    
    def add(self, pwd, cmds, *, priority=5, logpath=None):
        '''
        add [cmds]                    ls GPU task queue status
        
        Options:
            
            --priority [priority]     set task priority.
            --logpath  [path]         set task output file path.
        '''
        
        if logpath is None:
            logpath = os.path.join(self.logdir, f'{self._id_give}.log')
            
        task = Task(self, self._id_give, pwd, cmds, logpath)
        self._id_give += 1
        self.add_task(task, priority)
        
        result = f'add task(id:{task.id}) to queue(len: {len(self.queue)})'
        logging.info(f'add task(id:{task.id}): {pwd}$ {cmds})')
    
        err, result_ = self.scheduling.callback_add_process(self)
        
        return err, '\n'.join([result, result_])
    
    


task_manage = TaskManage(BaseScheduling())


@task_manage.client('ls')
def ls(*, all=False, sort='show'):
    '''
        ls                            ls GPU task queue status
    '''
    (all, sort), err_msg = check_input(((all, bool), (sort, str)))
    if err_msg: return 1, err_msg
    
#    print(sort)
    tasks = Sort.sort(task_manage.tasks, sort)
    if not isinstance(tasks, list): return 1, tasks
    
    table = pt.PrettyTable(['[ID]', 'num', 'status', 'run_times', 'pwd', 'cmds'])
    table.border = False
    for i, task in enumerate(tasks):
        status = str(task.status) + f'(GPU:{task.gpu})' if task.gpu is not None else str(task.status)
        if not all:
            table.add_row([task.id, i, status, task.run_times, task.pwd+'#', " ".join(task.cmds)[:80]])
        else:
            table.add_row([task.id, i, status, task.run_times, task.pwd+'#', " ".join(task.cmds)])
#    print(str(table))
    return 0, str(table)


@task_manage.client('rm')
def rm(id):
    '''
        rm [id]                       remove task [id] from manage, if task is running, kill it.
    '''
    (id,), err_msg = check_input(((id, int),))
    if err_msg:
        return 1, err_msg
    
    if task_manage.rm_task(id):
        return 0, f'[info]: del task {id}'
    else:
        return 1, f'[error]: can not found {id} in task queue.'


@task_manage.client('kill')
def kill(id):
    '''
        kill [id]                     kill task [id]
    '''
    
    (id,), err_msg = check_input(((id, int),))
    if err_msg:
        return 1, err_msg
    
    task = task_manage.get_task(id)
    if task is None:
        return 1, f'[error]: can not found id {id} in task queue.'
    return task.kill()
    

@task_manage.client('mv') 
def mv(id, index=0, *args, **kwargs):
    '''
        move [id] [index(default=0)]  move [id] to [index]
    '''
    (id, index), err_msg = check_input(((id, int),(index, int)), args, kwargs)
    if err_msg:
        return 1, err_msg
    
    if index > len(task_manage):
        return 2, f'[error]: index {index} is bigger than task queue length({len(task_manage)})'
    
    if task_manage.mv_task(id, index):
        return 0, f'[info]: move {id} to the first'
    else:
        return 1, f'[error]: can not found task {id}'


@task_manage.client('set') 
def set_property(name=None, value=None):
    '''
        set [name] [value]            set some property.
        
        Example:
            
            gpulimit set WAIT_TIME 1  set `WAIT_TIME=1`
    '''
    if name is None:
        return 0, '\n'.join([f'{k} = {v}' for k,v in task_manage.setter_param.items()])
    
    if name in task_manage.setter_param:
        if value is None:
            return 0, f'{name} = {task_manage.setter_param[name]}'
        
        value = int(value)
        task_manage.setter_param[name] = value
        result = 0, f'[info]: seted {name} = {value}'
    else:
        result = 1, f'[error]: name `{name}` can not set.'
    logging.info(result[1])
    return result
    

@task_manage.client('start') 
def start(id=None):
    '''
        start [iddefalut=None]        Force start task(s).
        
        Information:
            
            gpulimit start            running `check_and_start`, and auto start new task.
            gpulimit start 1          Force start task [id].
    '''
    
    if id is None:
        return task_manage.scheduling.user_start_scheduling(task_manage)
    (id, ), err_msg = check_input(((id, int),), )
    if err_msg:
        return 1, err_msg
    return task_manage.scheduling.user_start_scheduling(task_manage, id)


@task_manage.client('log')          
def get_output_filename(id):
    '''
        log [id]                      show [id] output.
        
        Example:
            
            gpulimit log 1            show task(id=1) output.
            gpulimit log main         show manage background log info.
    '''
    
    if id=='main':
        return 0, task_manage.log_file
    
    (id, ), err_msg = check_input(((id, int),), )
    if err_msg:
        return 1, err_msg
    
    task = task_manage.get_task(id)
    if task is None:
        return 1, 'Error'
    return 0, os.path.abspath(task.out_path)


@task_manage.client('status')   
def status(self):
    '''
        status                        show System status.
        
        Example:
            
            Nothing
    '''
    gpu_data = system_info.refresh().gpu
    
    task_nums = [0] * len(gpu_data)
    for task in self.queue:
        if task.gpu:
            task_nums[task.gpu] += 1
            
    table = pt.PrettyTable(['GPU[ID]', 'memory total', 'memory free', 'memory used', 'running tasks num'])
    table.border = False
    for info, use_num in zip(gpu_data, task_nums):
        table.add_row([info.id, info.memory_total, info.memory_free, info.memory_used, use_num])
        
    return 0, str(table)


@task_manage.client('debug')   
def debug(id):
    '''
        debug [id]                    if task [id] is `CMD_ERROR`, use this show error traceback.
        
        Example:
            
            debug 1                   show task 1 error traceback.                
    '''
    (id, ), err_msg = check_input(((id, int),), )
    if err_msg:
        return 1, err_msg
    
    task = task_manage.get_task(id)
    if task is None:
        return 1, f'[error]: can not found task[{id}]'
    
    return 0, str(task.debug_msg)


        



#! /usr/bin/python3

import sys, socket, os
import pickle as pk

parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__))) 
sys.path.insert(0, parentdir) 

from gpulimit.gpulimit_core.socket_utils import send_all, recv_all_str


if sys.platform == 'linux':
    server_address = '/tmp/gpulimit_uds_socket'
else:
    server_address = ('localhost', 5123)


def connect():
    if isinstance(server_address, str):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    else:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect(server_address)
    except socket.error:
        print('server not start.\n\n please use `gpulimit_server` to start server.')
        sys.exit(1)
    return sock


def show_help():
    print('help: this is help')


def main():
    if len(sys.argv) == 1:
        print(f'use `{sys.argv[0]} help` to show help message.')
        exit(0)
    sock = connect()
    pwd = os.getcwd()
    
    send_all(sock, pk.dumps([pwd, sys.argv[1:]]))
    result = recv_all_str(sock)
    sock.close()
    print(f'{result}')

    if sys.argv[1] == 'log':
        if result != 'Error':
            os.system(f'less {result}')
        else:
            print('[error]: please check task id.')


if __name__=='__main__':
    main()



# -*- coding: utf-8 -*-
import subprocess
import os, time, traceback
import threading
import logging


class TaskStatus(object):
    status2id = dict(zip(['CMD_ERROR', 'complete', 'waiting', 'running', 'runtime_error', 'killed'], range(-1, 5)))
    id2status = dict(zip(status2id.values(), status2id.keys()))
    
    can_start_list = ['waiting', 'runtime_error', 'killed']
    auto_start_list = ['waiting', 'runtime_error']
    
    def __init__(self, status='waiting', run_times=0, err_code=None):
        self._status = None
        self.run_times = run_times
        self.err_code = err_code
        
        if isinstance(status, int):
            self._set_status_id(status)
            
        elif isinstance(status, str):
            self._set_status_name(status)       
            
        else:
            raise TypeError(f'status must be int or str, not {type(status)}')
    
    @property
    def id(self):
        return self.status2id[self.status]
    
    @id.setter
    def id(self, nid):
        self._set_status_id(nid)
        
    @property
    def status(self):
        return self._status
    
    @status.setter
    def status(self, new_status):
        self._set_status_name(new_status)
        
    def _set_status_id(self, id):
            if id in self.id2status:
                self._status = self.id2status[id]
            else:
                raise ValueError(f'status id must be: {self.id2status}')
    
    def _set_status_name(self, name):
            if name in self.status2id:
                self._status = name
            else:
                raise ValueError(f'status name must be: {self.status2id}')
            
    def __str__(self):
        return self.status
    
    def __repr__(self):
        return f'TaskStatus({self.status})'


class Sort(object):
    start_sort_type = {
        'waiting': 0,
        'runtime_error': 1,
        'killed': 2,
        'running': 3,
        'complete': 3,
        'CMD_ERROR': 3,
    }
    show_sort_type = {
        'running': 1,
        'waiting': 2,
        'killed': 3,
        'runtime_error': 3,
        'complete': 4,
        'CMD_ERROR': 5,
    }
    
    sort_types = ['id', 'priority', 'show', 'run']
    
    def __call__(self, *args, **kwargs):
        return self.sort(*args, **kwargs)
    
    @staticmethod
    def sort(tasks, type):
        if type == 'run':
            tasks = sorted(tasks, key=lambda x: x.id)
            tasks = sorted(tasks, key=lambda x: Sort.start_sort_type[x.status.status])

        elif type == 'show':
            tasks = sorted(tasks, key=lambda x: x.id)
            tasks = sorted(tasks, key=lambda x: Sort.show_sort_type[x.status.status])
            
        elif type == 'id':
            tasks = sorted(tasks, key=lambda x: x.id)
            
        elif type == 'priority':
            tasks = sorted(tasks, key=lambda x: x.id)
            tasks = sorted(tasks, key=lambda x: x.priority)
        else:
            return f'[Error]: can not found sort type `{type}`, which can only use {Sort.sort_types}'

        return tasks
        
        
        
class Task(object):
    def __init__(self, task_manage, id, pwd, cmds, out_path=None):
        self.task_manage = task_manage
        self.id = id
        self.pwd = pwd
        self.cmds = cmds
        self.out_path = out_path
        
        self.priority = None
        self.pid = None
        self.gpu = None
        self.is_in_queue = False
        
        self.run_times = 0
        self.out_file = None
        self.pkg_process = None
        self.process = None
        self.available = True

        self.killed = False
        self.debug_msg = None
        
    def _run_task(self, GPU_id):
        try:
            if self.out_path is not None:
                self.out_file = open(self.out_path, 'w')
            self.gpu = GPU_id
            env = os.environ.copy()
            env['CUDA_VISIBLE_DEVICES'] = str(GPU_id)
            self.process = subprocess.Popen(self.cmds, stdout=self.out_file, shell=True, stderr=subprocess.STDOUT, cwd=self.pwd, env=env)
            self.pid = self.process.pid
            logging.info(f'[starting({self.id}: GPU:{GPU_id})]: {self.pwd}$ {self.cmds}')
        except Exception as e:
            self.available = False
            logging.info(f'[CMD_ERROR({self.id}: GPU:{GPU_id})]: {self.pwd}$ {self.cmds} \nerror:\n{e}')
            msg = traceback.format_exc()
            logging.info(msg)
            self.debug_msg = 'cmds: ' + str(self.cmds) + "\n" + msg
            self.task_manage.scheduling.callback_process_end(self.task_manage)
            return
        self.process.wait()
        self.gpu = None
        self.pid = None
        self.is_in_queue = False
        logging.info(f'[finish({self.id}: GPU:{GPU_id})]: {self.pwd}$ {self.cmds}')
        
        self.task_manage.scheduling.callback_process_end(self.task_manage)

    def start(self, GPU_id):
        if self.killed:
            self.killed = False
        if self.status.status in ['waiting', 'runtime_error', 'killed']:
            self.run_times += 1
            self.pkg_process = threading.Thread(target=self._run_task, args=(GPU_id,))
            self.pkg_process.start()
            
            return 0, f'[info]: start task {self.id} succeed.'
        else:
            return 1, f'[info]: can not start task {self.id} which have status `{self.status.status}`'
        
    def kill(self):
        if self.process is not None and self.status.status == 'running':
            self.process.terminate()
            time.sleep(1)
            self.process.kill()
            self.killed = True
            
            self.gpu = None
            self.pid = None
            self.is_in_queue = False
            return 0, f'[info]: kill task {self.id} succeed.'
        else:
            return 1, f'[warning]: can not kill task {self.id} which have status `{self.status.status}`'

    @property
    def status(self):
        if not self.available:
            return TaskStatus('CMD_ERROR', self.run_times)
        
        if self.process is None:
            return TaskStatus('waiting', self.run_times)
        
        if self.killed:
            return TaskStatus('killed', self.run_times)
        
        status = self.process.poll()
        if status is None:
            return TaskStatus('running', self.run_times)
        
        if status == 0:
            return TaskStatus('complete', self.run_times)
        else:
            return TaskStatus('runtime_error', self.run_times, err_code=status)


#!/usr/bin/env python3

from setuptools import setup, find_packages
import sys


with open("readme.md", "r", encoding='utf8') as fh:
    long_description = fh.read()


setup(
    name='gpulimit',
    version="0.1.0",
    author="lcy",
    author_email="lichunyang_1@outlook.com",
    description=("DL training on GPU management"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lichunown/gpu-limit",

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],

    packages=find_packages(),
    data_files=[],
    install_requires=[
        'prettytable',
    ],

    entry_points={'console_scripts': [
       'gpulimit-server = gpulimit.gpulimit_server:main',
       'gpulimit = gpulimit.gpulimit_client:main',
    ]},

    zip_safe=False
)